

local gps = require 'gps'
local ffi = require 'ffi'
local listener = require 'listener'
local dbus = require 'dbus'
local sample_clock = require 'sample_clock'
local bus = dbus.bus_get('system')
local timing = assert(bus:auto_proxy('com.somat.armada.Timing',
                                     '/com/somat/armada/Timing'))

local clkfreq = sample_clock.master_rate

local cmd_timeout = 0.01

local function cable_detect(devname)
    local gps_channel = gps:new(devname, "gps.bin", clkfreq)
    gps_channel:get_cable_detect_state()
    local msg = assert(listener({gps_channel}, cmd_timeout))
    local plug = {[0]="UNPLUGGED", [1]="PLUGGED"}
    print("GPS cable detect with " .. devname .. " is ",
          plug[msg.channel_specific[0]])
end

local function interactive_power_test(devname)
    local gps_channel = gps:new(devname, "gps.bin", clkfreq)
    io.write("Ready to turn on 5V") io.read()
    gps_channel:txdr_power_on_5v()
    assert(listener({gps_channel}, cmd_timeout))
    io.write("Ready to boost to 12V") io.read()
    gps_channel:txdr_power_on_12v()
    assert(listener({gps_channel}, cmd_timeout))
    io.write("Ready to turn off") io.read()
    gps_channel:txdr_power_off()
    assert(listener({gps_channel}, cmd_timeout))
    io.write("Off now") io.read()
    gps_channel:halt()
    print("Done")
    os.exit(0)
end

local function get_ts_test(devname)
    local gps_channel = gps:new(devname, "gps.bin", clkfreq)
    for i=1,5 do
        io.write("Press key to get current timestamp") io.read()
        gps_channel:get_current_ts()
        local msg = assert(listener({gps_channel}, cmd_timeout))
        print(msg.ts)
    end
    gps_channel:halt()
    print("Done")
    os.exit(0)
end

local function get_sync_test(devname)
    local gps_channel = gps:new(devname, "gps.bin", clkfreq)
    for i=1,5 do
        io.write("Press key to get latched timestamp") io.read()
        gps_channel:get_sync_ts()
        local msg = assert(listener({gps_channel}, cmd_timeout))
        print(msg.ts)
    end
    gps_channel:halt()
    print("Done")
    os.exit(0)
end

local function synchronization_test(devname)
    local gps_channel = gps:new(devname, "gps.bin", clkfreq)
    gps_channel:setup_sync_ts()
    assert(listener({gps_channel}, cmd_timeout))
    local sync_time = timing:SendSyncPulse()
    gps_channel:get_sync_ts()
    local msg = assert(listener({gps_channel}, cmd_timeout))
    local offset = sync_time - msg.ts + 1
    gps_channel:ts_offset(offset)
    assert(listener({gps_channel}, cmd_timeout))
    gps_channel:setup_sync_ts()
    assert(listener({gps_channel}, cmd_timeout))
    sync_time = timing:SendSyncPulse()
    gps_channel:get_sync_ts()
    msg = assert(listener({gps_channel}, cmd_timeout))
    offset = sync_time - msg.ts
    if offset == 0 then
        print("Passed")
    else
        print("Failed, offset is", offset)
    end
    gps_channel:halt()
    print("Done")
    os.exit(0)
end

local function display_heartbeat_messages(channel_obj, name, ts)
    io.write("HB ") io.flush()
    channel_obj.heartbeat_count = channel_obj.heartbeat_count + 1
end

local function heartbeat_test(devname)
    local gps_channel = gps:new(devname, "gps.bin", clkfreq)
    gps_channel.heartbeat_func = display_heartbeat_messages
    gps_channel.heartbeat_count = 0
    gps_channel:get_current_ts()
    local msg = assert(listener({gps_channel}, cmd_timeout))
    local start_time = msg.ts + clkfreq
    gps_channel:event_start(start_time - 3, 4, 16, 128)
    assert(listener({gps_channel}, cmd_timeout))
    listener({gps_channel}, 10.0)
    print("\nCleaning up")
    gps_channel:event_stop()
    assert(listener({gps_channel}, cmd_timeout))
    gps_channel:halt()
    print("Heartbeats detected = "..gps_channel.heartbeat_count)
    print("Done")
    os.exit(0)
end

local function local_loopback_test(devname)
    local capture_log = ""
    local gps_channel = gps:new(devname, "gps.bin", clkfreq)
    gps_channel.element_action_func = function(channel_obj, name, pktstruct)
        capture_log = capture_log..string.char(pktstruct.rxdata)
    end
    gps_channel:get_current_ts()
    local msg = assert(listener({gps_channel}, cmd_timeout))
    local start_time = msg.ts + clkfreq
    gps_channel:uart_rx_start(start_time - 3, 2, 16, 128, '8N1', 38400)
    assert(listener({gps_channel}, cmd_timeout))
    gps_channel:uart_tx_start(10)
    assert(listener({gps_channel}, cmd_timeout))

    listener({gps_channel}, 0.1)
    gps_channel:uart_tx_data("Hello")
    listener({gps_channel}, 0.1)
    gps_channel:uart_tx_data("World")
    listener({gps_channel}, 2.0)

    print("Cleaning up")
    gps_channel:uart_tx_stop()
    assert(listener({gps_channel}, cmd_timeout))
    gps_channel:uart_rx_stop()
    assert(listener({gps_channel}, cmd_timeout))
    gps_channel:halt()

    if capture_log == "HelloWorld" then print("Passed") else print("Failed") end
    print("Done")
    os.exit(0)
end

local function gps_monitor_test(devname)
    local capture_log = {}
    local pps_rise_log = {}
    local pps_fall_log = {}
    local start_time
    capture_log.dropped_characters = ""
    local gps_channel = gps:new(devname, "gps.bin", clkfreq)
    gps_channel.element_action_func = function(channel_obj, name, pktstruct)
        if name == 'EV' then
            if pktstruct.events.pps_rising == 1 then
                pps_rise_log[#pps_rise_log + 1] = pktstruct.ts
            end
            if pktstruct.events.pps_falling == 1 then
                pps_fall_log[#pps_fall_log + 1] = pktstruct.ts
            end
        else
            local c = string.char(pktstruct.rxdata)
            if c == '$' then
                local new_nmea_record = {}
                new_nmea_record.s = c
                new_nmea_record.ts = pktstruct.ts - start_time
                capture_log[#capture_log + 1] = new_nmea_record
            else
                if #capture_log > 0 then
                    local nmea_record = capture_log[#capture_log]
                    nmea_record.s = nmea_record.s..c
                else
                    capture_log.dropped_characters =
                        capture_log.dropped_characters..c
                    io.write(c)
                end
            end
        end
    end
    gps_channel:txdr_power_on_5v()
    assert(listener({gps_channel}, cmd_timeout))
    gps_channel:event_setup({pps_falling = true, pps_rising = true})
    assert(listener({gps_channel}, cmd_timeout))
    gps_channel:get_current_ts()
    local msg = assert(listener({gps_channel}, cmd_timeout))
    start_time = msg.ts + clkfreq
    gps_channel:uart_rx_start(start_time - 3, 2, 16, 128, '8N1', 38400)
    assert(listener({gps_channel}, cmd_timeout))
    gps_channel:event_start(start_time - 3, 2, 16, 128)
    assert(listener({gps_channel}, cmd_timeout))

    io.write("Waiting for PPS lock")
    local waiting_for_lock = true
    local end_time = 120
    local current_time = 0
    while current_time < end_time do
        listener({gps_channel}, 1.0)
        if #pps_rise_log > 0 and waiting_for_lock then
            print("\nPPS lock achieved.")
            io.write("Collecting 30 seconds of pps data")
            waiting_for_lock = false
            end_time = current_time + 30
        end
        current_time = current_time + 1
        io.write('.') io.flush()
    end

    print("\nCleaning up")
    gps_channel:uart_rx_stop()
    assert(listener({gps_channel}, cmd_timeout))
    gps_channel:event_stop()
    assert(listener({gps_channel}, cmd_timeout))
    gps_channel:halt()
    print("Done")
    print("Captured GPS Data")
    local drops = #capture_log.dropped_characters
    if drops > 0 then
        print("Dropped "..drops.." characters before find a '$' marker.")
    end
    local total = #capture_log
    print("Total NMEA sentences captured = "..total)
    local to_show = 5
    if total < 6 then to_show = total - 1 end
    for i=total-5,total-1 do
        io.write(capture_log[i].s)
    end
    local local_transitions = #pps_rise_log + #pps_fall_log
    print("Total timestamp transitions = "..local_transitions)
    local compute_average = function(ts_log)
        local avg = 0
        for i=1,#ts_log-1 do avg = avg + (ts_log[i+1] - ts_log[i]) end
        return tonumber(avg) / (#ts_log - 1)
    end
    print("Average rate of PPS rising edges = "..
            clkfreq / compute_average(pps_rise_log).."Hz")
    print("Average rate of PPS falling edges = "..
            clkfreq/ compute_average(pps_fall_log).."Hz")
    os.exit(0)
end




if #arg < 1 then
    print("Usage gps_test <device name> [POWER_TEST|LOCAL_LOOPBACK|GET_TS" ..
          "|GET_SYNC|SYNC_TEST|CABLE_DETECT]")
    os.exit(1)
end

if #arg > 1 then
    if arg[2] == "CABLE_DETECT" then cable_detect(arg[1]) end
    if arg[2] == "POWER_TEST" then interactive_power_test(arg[1]) end
    if arg[2] == "LOCAL_LOOPBACK" then local_loopback_test(arg[1]) end
    if arg[2] == "GET_TS" then get_ts_test(arg[1]) end
    if arg[2] == "GET_SYNC" then get_sync_test(arg[1]) end
    if arg[2] == "SYNC_TEST" then synchronization_test(arg[1]) end
    if arg[2] == "HEARTBEAT_TEST" then heartbeat_test(arg[1]) end
    print("Unrecognized test name")
    os.exit(1)
end

gps_monitor_test(arg[1])
