local ffi = require 'ffi'
local message_can = require 'message_can'
local pcre = require 'pcre_ffi'
local sample_clock = require 'sample_clock'

local C = ffi.C

local ffi_copy = ffi.copy
local ffi_new = ffi.new
local ffi_sizeof = ffi.sizeof
local ffi_string = ffi.string
local string_format = string.format

ffi.cdef(string_format("typedef int%d_t ssize_t", 8 * ffi_sizeof'size_t'))

ffi.cdef[[
    typedef int64_t timestamp;
    typedef float float32_t;
    typedef double float64_t;

    typedef struct __attribute((packed)) {
        uint32_t length;
    } proxy_length_header;

    typedef struct __attribute((packed)) {
        uint32_t length;
        timestamp timestamp[1];
    } proxy_header;

    /* Sample types */

    /* Fixed-size sequential data types algorithmically defined */

    /* Fixed-size message data */

    typedef struct __attribute((packed)) {
        timestamp timestamp[1];
        float value[1];
    } message_float32;

    typedef struct __attribute((packed)) {
        timestamp timestamp[1];
        float value[1];
    } message_float32_ntp;

    typedef struct __attribute((packed)) {
        uint32_t length;
        message_float32 sample[1];
    } message_float32_buf;

    typedef struct __attribute((packed)) {
        timestamp timestamp[1];
        double value[1];
    } message_float64;

    typedef struct __attribute((packed)) {
        timestamp timestamp[1];
        double value[1];
    } message_float64_ntp;

    typedef struct __attribute((packed)) {
        uint32_t length;
        message_float64 sample[1];
    } message_float64_buf;

    typedef struct __attribute((packed)) {
        uint32_t length;
        message_can sample[1];
    } message_can_buf;

    /* Variable-size message data */

    typedef struct __attribute((packed)) {
        uint32_t length;
        timestamp timestamp[1];
        uint8_t value[1];
    } message_binary_header;

    typedef struct __attribute((packed)) {
        uint32_t length;
        timestamp timestamp[1];
        uint8_t sample[1];
    } message_raw_buf;

    typedef struct __attribute((packed)) {
        uint32_t length;
        timestamp timestamp[1];
        uint8_t *value;
    } message_raw;
]]

local NaN = 0/0

local function isnan(x)
    return x ~= x
end

local type_mt = {}

local hex = ffi.new('uint8_t[16]', '0123456789ABCDEF')

local function extend(t, ...)
    for _, nt in ipairs{...} do
        for k, v in pairs(nt) do
            t[k] = v
        end
    end
    return t
end

local base_mt = {}

function base_mt:byte_length()
    return self:is_heartbeat() and 0 or ffi_sizeof(self.value[0])
end

function base_mt:__len()
    return self:byte_length()
end

function base_mt:is_heartbeat()
    return false
end

function base_mt:set_heartbeat()
    return false
end

function base_mt:dump_hex()
    local size = ffi_sizeof(self)
    local buf = ffi_new('uint8_t[?]', size * 4, 0x20)
    local pos = 0
    local cp = ffi_new('uint8_t[?]', size)
    ffi_copy(cp, self, size)
    for i = 0, size - 1 do
        buf[pos] = hex[bit.band(0x0F, bit.rshift(cp[i], 4))]
        buf[pos + 1] = hex[bit.band(0x0F, cp[i])]
        pos = pos + 2
        if i == 7 then
            pos = pos + 1
        end
    end
    return string_format("%s (%p %s)",
        tostring(self), self, ffi_string(buf, pos))
end

function base_mt:dump()
    return self:dump_hex()
end

-- sequential_* types

local seq_cdef = [[
    typedef struct __attribute((packed)) {
        $c value[1];
    } $t;

    typedef struct __attribute((packed)) {
        uint32_t length;
        timestamp timestamp;
        $t sample[1];
    } $t_buf;
]]

for _, pfx in ipairs{'float', 'int', 'uint'} do
    for _, size in ipairs{8, 16, 32, 64} do
        if pfx ~= 'float' or size >= 32 then
            local map = {}
            map.t = 'sequential_' .. pfx .. size
            map.c = pfx .. size .. '_t'
            local cdef = seq_cdef:gsub('%$(%w)', map)
            ffi.cdef(cdef)
            type_mt[map.t] = base_mt
        end
    end
end

-- message_float* types

local mf_mt = extend({}, base_mt)
type_mt.message_float32 = mf_mt
type_mt.message_float64 = mf_mt

function mf_mt.is_heartbeat(msg)
    return isnan(msg.value[0])
end

function mf_mt.set_heartbeat(msg, setting)
    msg.value[0] = setting and NaN or msg.value[0]
end

function mf_mt:dump()
    return string_format('%.9f: %f',
        sample_clock.secs(self.timestamp[0]), tonumber(self.value[0]))
end

function mf_mt:__tostring()
    return string_format('%f', tonumber(self.value[0]))
end

-- message_raw type

local raw_mt = extend({}, base_mt)
type_mt.message_raw = raw_mt

function raw_mt:byte_length()
    return self.length - ffi_sizeof'timestamp'
end

function raw_mt:is_heartbeat()
    return self:byte_length() == 0
end

function raw_mt:set_heartbeat(setting)
    self.length = ffi_sizeof'timestamp'
end

function raw_mt:dump()
    return string_format('%.9f: %s',
        sample_clock.secs(self.timestamp[0]), ffi_string(self.value, #self))
end

function raw_mt:__tostring()
    return ffi_string(self.value, #self)
end

-- Make all the metatypes

local ffi_metatypes = {}
for type_name, mt in pairs(type_mt) do
    mt.__index = mt
    ffi_metatypes[type_name] = ffi.metatype(type_name, mt)
end

local M = {}

function M:constructor(type_name)
    return ffi_metatypes[type_name]
end

function M:new(type_name, ...)
    local r = (self:constructor(type_name))(...)
end

local has_heartbeats_rx = pcre:new('^message_(can|float\\d+)$')
local is_fixed_rx = pcre:new('^((sequential|message)_(can|(u?int|float)\\d+))$')
local is_numeric_rx = pcre:new('^((sequential|message)_(u?int|float)\\d+)$')

assert(is_fixed_rx:match('sequential_float32'))
assert(is_fixed_rx:match('sequential_float64'))
assert(is_fixed_rx:match('message_float32'))
assert(is_fixed_rx:match('message_float64'))
assert(is_fixed_rx:match('message_can'))

function M.data_class(data_type)
    return M.is_message(data_type) and 'message' or 'sequential'
end

function M.has_heartbeats(data_type)
    return not not has_heartbeats_rx:match(data_type or '')
end

function M.is_fixed_size(data_type)
    return not not is_fixed_rx:match(data_type or '')
end

function M.is_message(data_type)
    return not not (data_type or ''):match('^message_')
end

function M.is_numeric(data_type)
    return not not is_numeric_rx:match(data_type or '')
end

function M.is_sequential(data_type)
    return not not (data_type or ''):match('^sequential_')
end

function M.is_integer(data_type)
    return not not (data_type or ''):match('_[u]?int%d+$')
end

function M.bits(data_type)
    return tonumber(data_type:match('(%d+)$')) or 0
end

function M.to_float_type(data_type)
    if M.is_numeric(data_type) then
        local bits = tonumber(data_type:match('(%d+)$'))
        bits = (bits <= 32) and 32 or 64
        return M.is_sequential(data_type)
            and 'sequential_float' .. bits
            or 'message_float' .. bits
    end
end

function M.to_sequential_type(data_type)
    if data_type:match('^sequential_') then
        return data_type
    elseif M.is_numeric(data_type) and M.is_fixed_size(data_type) then
        local bits = tonumber(data_type:match('(%d+)$'))
        bits = (bits <= 32) and 32 or 64
        local parts = data_type:match('^sequential_(float)')
        if not parts then
            parts = data_type:match('^sequential_(u?int)')
        end
        if not parts then
            parts = data_type:match('^message_(float)')
        end
        if not parts then
            parts = data_type:match('^message_(u?int)')
        end
        return 'sequential_' .. parts .. bits
    else
        return nil
    end
end

return M
