-- Implements an asynchronous task chain pattern which executes a sequence
-- of functions, each of which calls a callback (passed as an argument to
-- the task function) when it has completed. Tasks can be chained together
-- by passing subsequent task functions as callbacks to previous tasks. This
-- pattern is common in Armada code, and this class provides a convenient
-- and readable way to implement it. See the Wiki for more documentation on
-- this design pattern and how to use this class.
--
-- To prevent inadvertent race conditions, task chains may not be modified
-- after run() has been called. Also, run() may only be called once for a
-- given task chain object.
--
-- NOTE: This module is NOT thread-safe. Calling any task_chain functions from
-- multiple threads simultaneously results in undefined behavior. Task
-- functions themselves CAN safely use threads, so long as they do not call
-- any task_chain functions directly.
--
-- Brief example:
--
--     local task_chain = require 'task_chain'
--     local tasks = task_chain:new()
--     tasks:add(start_task1)
--     tasks:add(start_task2)
--     tasks:run(on_success, on_error)
--
-- Task start function example:
--
--     function start_task1(on_success, on_error)
--         -- Process task in background
--         -- Call on_success when task is completed
--     end

require 'middleclass'

local task_chain = class('task_chain')

-- A local helper function to recursively walk the chain of tasks. The
-- first queued task function is called, with this function as the task's
-- completion callbacks. When the callbacks are triggered, this function
-- calls itself to process the next task.

local function _run_all_tasks(task_list, success_cb, error_cb)

    local n = #task_list

    if task_list and #task_list > 0 then

        local next_task = table.remove(task_list, 1)

        local on_success = function()
            _run_all_tasks(task_list, success_cb, error_cb)
        end

        local on_error = function(message)
            message = message or 'Unknown error'
            if error_cb then error_cb(message) end
        end

        next_task(on_success, on_error)

    else

        if success_cb then success_cb() end

    end

end

-- Adds a new task to the end of the list of tasks to perform. The 'task_func'
-- parameter is a function of the form task_func(success_cb, error_cb). The
-- function can return immediately while doing processing in the background.
-- When the task is complete, either 'success_cb' or 'error_cb' should be
-- called, depending on whether the task succeeded or not. 'success_cb'
-- should be called with no arguments, while 'error_cb' may optionally be
-- called with a single string error message argument.
--
-- If this function is called after run() has been called, it will do
-- nothing (the task will not be added). Task chains cannot be modified
-- after run() is called.

function task_chain:add(task_func)
    if self.started == 0 then
        self.tasks[#self.tasks + 1] = task_func
    end
end

-- Runs all tasks in this chain. This function may return immediately. If
-- all tasks complete successfully, 'success_cb' is called with no arguments.
-- If an error occurred (one of the queued task functions called its
-- 'error_cb' argument), 'error_cb' is called with a single string error
-- message argument.
--
-- run() may only be called once for a given task chain. Subsequent calls
-- result in error_cb() being called immediately with an error message.

function task_chain:run(success_cb, error_cb)
    if self.started ~= 0 then
        if error_cb then error_cb("task_chain:run() already called") end
    else
        self.started = 1
        _run_all_tasks(self.tasks, success_cb, error_cb)
    end
end

-- Initializes an empty task chain.

function task_chain:initialize()
    self.tasks = {}
    self.started = 0
end

return task_chain
