require 'strict'

-- burst_history.lua - Implements the Burst History data mode.
--
-- Implementation notes:
--     This data mode uses its own particular trigger mechanism,
--     not the "generic" datamode triggering mechanism.
--
--     The "Storage mode" option "Standard" works as Burst History
--     has always worked, CX23-R or eDAQ:  triggers are not recognized
--     during collection of a burst, and all bursts stored have the
--     same size (their SIE file blocks are all the same size).
--
--     The "Storage mode" option "Extended" recognizes all triggers
--     whenever they occur and guarantees that at least a
--     'burst_post_trigger_time' amount of burst history is recorded
--     after each trigger (with the possible exception of the end of
--     the test, of course).  This implies that bursts will be extended
--     past the standard "pre-count + post-count + 1 sample" size,
--     though these extensions take the form of contiguous blocks
--     after the standard first block.  So burst storage blocks are not
--     guaranteed to have the same size, and probably won't if a trigger
--     occurs while storing a burst.  Also, trigger offsets will be
--     positive for all samples in burst extension blocks.
--
--     Sample counts are used throughout, and stored using simple Lua
--     floating-point number variables. This is safe. IEEE 754 double-
--     precision arithmetic suffers no loss of precision or rounding errors
--     as long as operands and results stay below 2^53 - 1. A channel
--     sampled at 10 MHz (well above anything supported by the CX23-R/eDAQXR)
--     will not reach this sample count until (2^53-1)/1e7 seconds, or about
--     28.56 years.

local ffi               = require 'ffi'
local frame             = require 'test_engine/frame'
local pipe              = require 'test_engine/pipe'
local rational          = require 'rational'
local sie_schema        = require 'test_engine/sie_schema'
local sie_utils         = require 'test_engine/sie_utils'
local test_eng_utils    = require 'test_engine/test_eng_utils'

local m                 = {}
local mt                = {}

local EPSILON           = 0.000001

-------------------------------------------------------------------------------

function m:new(spec, nodes, first_sie_index, metas)

    -- Create the new object and save off the SXR definition and SIE group #.
    local bh = setmetatable({}, { __index = mt })
    bh.spec = spec
    bh.prefix_dm_name = bh.spec.summary.include_dm_name_and_at

    -- Extract some useful information out of the SXR definition.
    local sample_rate = rational:new(metas[spec.input_channels[1]].sample_rate)
    bh.pre_trigger_count = math.floor(sample_rate:tonumber() *
        spec.burst_pre_trigger_time)
    bh.post_trigger_count = math.floor(sample_rate:tonumber() *
        spec.burst_post_trigger_time)
    bh.enhanced_mode = spec.storage_mode == 'enhanced'

    -- Pick the appropriate trigger function depending on whether we're
    -- configured as edge-triggered or level-triggered.
    bh.trigger_func = ({
        when_true = bh._level_trigger,
        on_false_to_true_edge = bh._edge_trigger
    })[spec.trigger_mode]

    -- Create some general bookkeeping variables.
    bh.burst_start_ts   = nil
    bh.burst_until_ts   = nil
    bh.last_trigger     = false

    -- Create a data structure to keep track of the trigger pipe data.
    local trigger_node = nodes[spec.trigger_channel]
    bh.trigger = {
        pipe                = pipe.attach_pipe(trigger_node.pipe_list),
        consumed_from_head  = 0,
        first_unread_ts     = 0
    }

    -- Create data structures for each data channel to keep track of
    -- their input and output pipes and timestamps.
    bh.channels = {}
    for i = 1, #spec.input_channels do
        local node = nodes[spec.input_channels[i]]
        bh.channels[i] = {
            spec                = metas[spec.input_channels[i]],
            in_pipe             = pipe.attach_pipe(node.pipe_list),
            out_pipe_list       = pipe.new_pipe_list(),
            consumed_from_head  = 0,
            first_unread_ts     = 0,
            sie_buf_info = {
                idx = spec.sie_buf_info.idx + i - 1,
                id = spec.sie_buf_info.id + i - 1,
                group = spec.sie_buf_info.group + i - 1,
            },
        }
    end

    return bh, #spec.input_channels

end

-------------------------------------------------------------------------------

function mt:setup(sie_writer)

    self.sie_writer = sie_writer

    for i = 1, #self.channels do

        local channel = self.channels[i]

        channel.spec.description = channel.spec.description or ''
        channel.spec.datamode_name = self.spec.name
        if self.prefix_dm_name then
            channel.spec.sie_name = self.spec.name .. '@' .. channel.spec.name
        else
            channel.spec.sie_name = channel.spec.name
        end
        channel.somat_input_channel = channel.spec.name

        self.spec.nbr_data_bits = tonumber(
            channel.spec.data_type:match('^.-(%d+)$'))

        local schema_name = sie_schema.get_dm_schema_name(self.spec.category)
        local schema_info = sie_schema.get_schema_info(schema_name, self.spec)
        schema_info.dims[2]['core:label'] = channel.spec.name
        schema_info.dims[2]['core:units'] = channel.spec.output_units

        channel.spec.sie_buf_info = channel.sie_buf_info
        local md = sie_utils.write_sie_channel(
            self.sie_writer, schema_info, channel.spec)
        local p = pipe.attach_pipe(channel.out_pipe_list)
        local size = test_eng_utils.sie_buf_size(self, channel.spec)
        self.sie_writer:new_buffer(channel.sie_buf_info.idx, 
            channel.sie_buf_info.group, size,
            ffi.sizeof(channel.spec.data_type), 8,
            p, nil, sie_utils.BH_DATAMODE)

    end

end

-------------------------------------------------------------------------------

function mt:_edge_trigger(value)
    local set = math.abs(value) >= EPSILON
    local last = self.last_trigger_value
    self.last_trigger_value = set
    return set and not last
end

function mt:_level_trigger(value)
    return math.abs(value) >= EPSILON
end

-------------------------------------------------------------------------------

-- Searches for either the first or the last triggered sample in the range
-- from the end of the previous search up to but excluding end_ts. If a
-- triggered sample is found, its timestamp is returned and all samples up
-- to and including the triggered sample are consumed from the trigger pipe.
-- Otherwise, nil is returned and all samples up to but excluding end_ts are
-- consumed. The caller must ensure that end_ts is never less than any
-- previous end_ts.

function mt:_find_trigger(end_ts, want_first, want_last)

    local trigger_ts = nil

    while end_ts > self.trigger.first_unread_ts do

        local head_frame = self.trigger.pipe.last_frame

        local n = math.min(end_ts - self.trigger.first_unread_ts,
            head_frame.count - self.trigger.consumed_from_head)

        if want_first or want_last then

            local values = ffi.cast(
                test_eng_utils.frame_data_type(head_frame.element_size),
                head_frame.buf) + self.trigger.consumed_from_head

            for i = 0, n - 1 do
                if self:trigger_func(values[i]) then
                    trigger_ts = self.trigger.first_unread_ts + i
                    if not want_last then
                        end_ts = trigger_ts + 1
                        n = i + 1
                        break
                    end
                end
            end

        end

        if n == head_frame.count - self.trigger.consumed_from_head then
            frame.delete(pipe.get_frame(self.trigger.pipe))
            self.trigger.consumed_from_head = 0
        else
            self.trigger.consumed_from_head =
                self.trigger.consumed_from_head + n
        end

        self.trigger.first_unread_ts = self.trigger.first_unread_ts + n

    end

    return trigger_ts

end

-------------------------------------------------------------------------------

-- Dumps all samples before a specified timestamp, optionally pushing them to
-- the output (if 'trigger_offset' is not nil). Fully-consumed frames are
-- removed from the input pipe and destroyed. Each channel's first_unread_ts
-- value is advanced to end_ts.

function mt:_dump_input(end_ts)

    -- Keep track of the total number of samples processed for the first
    -- channel (whether or not actually written to SIE). We only care about
    -- the first channel; the current algorithm guarantees that all channels
    -- will have the same first_unread_ts value, so the result is the same
    -- for all channels anyway.
    local written = nil

    for i = 1, #self.channels do

        local channel = self.channels[i]
        local written_this_channel = 0

        while end_ts > channel.first_unread_ts do

            local head_frame = channel.in_pipe.last_frame

            local n = math.min(end_ts - channel.first_unread_ts,
                head_frame.count - channel.consumed_from_head)

            if self.burst_start_ts then

                if n == head_frame.count then

                    -- Optimization: If the entire frame is to be written,
                    -- we can simply clone it (avoiding having to copy its
                    -- data) and send that to the output pipe.
                    local out_frame = frame.clone(head_frame)
                    out_frame.offset = channel.first_unread_ts -
                        self.burst_start_ts
                    if not test_eng_utils.put_frame(out_frame,
                            channel.out_pipe_list) then
                        frame.delete(out_frame)
                    end

                else

                    local out_frame = frame.create(n, head_frame.element_size)
                    out_frame.hdr.ts = head_frame.hdr.ts +
                        channel.consumed_from_head
                    out_frame.ts = channel.first_unread_ts
                    out_frame.offset = channel.first_unread_ts -
                        self.burst_start_ts
                    out_frame.count = n
                    ffi.copy(out_frame.buf,
                        ffi.cast('char *', head_frame.buf) +
                            head_frame.element_size *
                            channel.consumed_from_head,
                        n * head_frame.element_size)
                    if not test_eng_utils.put_frame(out_frame,
                            channel.out_pipe_list) then
                        frame.delete(out_frame)
                    end

                end

            end

            channel.consumed_from_head = channel.consumed_from_head + n
            channel.first_unread_ts = channel.first_unread_ts + n

            if channel.consumed_from_head == head_frame.count then
                frame.delete(pipe.get_frame(channel.in_pipe))
                channel.consumed_from_head = 0
            end

            written_this_channel = written_this_channel + n

        end

        if not written then written = written_this_channel end

    end

    return written

end

-------------------------------------------------------------------------------

function mt:run()

    -- Get the latest timestamp at which we all data has arrived.
    local process_until = self.trigger.first_unread_ts +
        tonumber(self.trigger.pipe.sample_count) -
        self.trigger.consumed_from_head
    for i = 1, #self.channels do
        local channel = self.channels[i]
        local ts = channel.first_unread_ts +
            tonumber(channel.in_pipe.sample_count) -
            channel.consumed_from_head
        if ts < process_until then process_until = ts end
    end

    while process_until > self.trigger.first_unread_ts do

        if self.burst_start_ts then

            local horizon = math.min(self.burst_until_ts, process_until)

            -- If we're in enhanced mode, look for any re-triggers in the
            -- existing burst window, and if found, extend the burst window.
            -- Otherwise, ignore all triggers.
            if self.enhanced_mode then
                local t = self:_find_trigger(horizon, false, true)
                if t then
                    self.burst_until_ts = t + self.post_trigger_count + 1
                end
            else
                self:_find_trigger(horizon)
            end

            -- Dump the data.
            self:_dump_input(horizon)

            -- If we've reached the end of the burst, update accordingly.
            if horizon == self.burst_until_ts then
                self.burst_start_ts = nil
                self.burst_until_ts = nil
                self.sie_writer:flush_these(self.sie_collect_idx,
                    #self.channels, true)
            end

        else

            local t = self:_find_trigger(process_until, true)

            if t then

                -- New burst starting. Discard any data prior to the first
                -- pre-trigger sample.
                self:_dump_input(t - self.pre_trigger_count)

                -- Initialize the burst bookkeeping variables, and then
                -- dump out the pre-trigger data, if any.
                self.burst_start_ts = t
                self.burst_until_ts = t + self.post_trigger_count + 1
                local written = self:_dump_input(t + 1)

                -- (#6122) If the available pre-trigger data is less than
                -- requested, and we're in standard recording mode, extend the
                -- burst so that its total size will be exactly equal to the
                -- requested size.
                if not self.enhanced_mode then
                    self.burst_until_ts = self.burst_until_ts +
                        (self.pre_trigger_count + 1 - written)
                end

            else

                -- Not in a burst, and no trigger occurred in the available
                -- data. Dump any input data older than the pre-trigger
                -- sample count because we're guaranteed not to need it.
                self:_dump_input(process_until - self.pre_trigger_count)

            end

        end

    end

end

-------------------------------------------------------------------------------

function mt:finalize(sie_writer)
end

-------------------------------------------------------------------------------

return m
