require 'strict'

-- event_slice.lua - Implements the Event Slice data mode.
--
local cjson = require 'cjson'
local ffi = require 'ffi'
local frame = require 'test_engine/frame'
local __ = require '__'
local log = require 'log'
local pipe = require 'test_engine/pipe'
local rational = require 'rational'
local sie_schema = require 'test_engine/sie_schema'
local sie_utils = require 'test_engine/sie_utils'
local utils = require 'test_engine/test_eng_utils'

local ffi_cast = ffi.cast
local ffi_sizeof = ffi.sizeof

local Frame_create = frame.create
local Frame_delete = frame.delete
local Frame_tail   = frame.tail
local Pipe_attach_pipe   = pipe.attach_pipe
local Pipe_has_frames    = pipe.has_frames
local Pipe_get_frame     = pipe.get_frame
local Pipe_new_pipe_list = pipe.new_pipe_list


local mt = {}


-- Performs any needed post instantiation initialization.
-- Parameter:
--     'sie_writer'  SIE writer
-- Returns:  true
function mt:setup(sie_writer)
    self.sw = sie_writer
    self:sie_xml()
    return true
end

-- Creates the SIE buffers (pipes and XML) required.
-- There will be one channel description per input channel.
function mt:sie_xml()
    for i = 1, self.nbr_output_chans do
        local info = self.chan_info[i]
        -- write XML for this SIE channel
        info.stanza.description = self.stanza.description or ""
        info.stanza.datamode_name = self.name
        if self.prefix_dm_name then
            info.stanza.sie_name = self.name .. '@' .. info.name
        else
            info.stanza.sie_name = info.name
        end
        info.stanza.somat_input_channel = info.stanza.name
        -- the following line makes the schema info reflect the
        -- input channel's data bit size (?should this be different?)
        self.stanza.nbr_data_bits =
            tonumber(info.data_type:match('^.-(%d+)$'))
        assert(self.stanza.nbr_data_bits and self.stanza.nbr_data_bits > 0,
            "Missing data bit size")
        local schema_name = sie_schema.get_dm_schema_name(self.stanza.category)
        local schema_info =
            sie_schema.get_schema_info(schema_name, self.stanza)
        schema_info.somat_datamode_type = 'event_slice'
        schema_info.dims[1]['core:label'] = "Event Count"
        schema_info.dims[1]['core:units'] = nil
        schema_info.dims[2]['core:label'] = info.stanza.name
        schema_info.dims[2]['core:units'] = info.stanza.output_units
        schema_info.xform_scale = 1.0
        schema_info['core:uuid'] = self.stanza.id .. '/' .. info.stanza.id

        info.stanza.sie_buf_info = info.sie_buf_info
        local md = sie_utils.write_sie_channel(self.sw, schema_info,
            info.stanza)
        -- Attach one end of pipe to this module for this input node
        local p = Pipe_attach_pipe(info.pipe_list)
        -- Create the SIE buffer and attach the pipe to it
        local size = utils.sie_buf_size(self, info.stanza)
        self.sw:new_buffer(info.sie_buf_info.idx, info.sie_buf_info.group,
            size, ffi_sizeof(info.data_type), 8, p)
    end
end

-- Gets a frame from each input pipe, storing the frames in
-- the input pipe's 'self.chan_info[<pipe index>].in_frame' entry.
-- Returns:  true if a full set of frames is acquired,
--           false otherwise
function mt:get_frames()
    -- check that each input pipe has a frame, so we don't
    -- "get" a frame from any pipe which we can't use
    local got_frames = false

    for i = 1, #self.chan_info do
        got_frames = Pipe_has_frames(self.chan_info[i].in_pipe)
        if not got_frames then
            break  -- don't have a frame from this input pipe
        end
    end
    if got_frames then
        for i = 1, #self.chan_info do
            self.chan_info[i].in_frame =
                Pipe_get_frame(self.chan_info[i].in_pipe)
        end
    else
        -- ensure we don't accidentally use old data
        for i = 1, #self.chan_info do
            self.chan_info[i].in_frame = nil
        end
    end
    return got_frames
end

-- Checks the master input channels for an event.  An event in any
-- master channel will do.  An event is defined to be any sequential
-- change in the channels data sample values.
-- Parameter:
--     'sample_index'  index of sample(s) in master input channel(s)
--         data to check
-- Returns:  true if an event was found, false otherwise
function mt:check_for_event(sample_index)
    local is_event = false
    local info
    for ch_index = 1, self.nbr_master_chans do
        info = self.chan_info[ch_index]
        if info.invals[sample_index] ~= info.prev_state then
            info.prev_state = info.invals[sample_index]
            is_event = true
        end
    end
    return is_event
end

-- Processes incoming frames, producing outgoing frames.
function mt:run()
    if self.trigger_module then
        -- trigger node feeds this node's input channels
        self.trigger_module:run()
    end

    -- attempt to get a full set of input frames (one from each
    -- input pipe)
    local have_all_frames = self:get_frames()
    if not have_all_frames then
        -- don't have a full set of input frames yet
        return
    end

    -- Process each frame in input pipe(s).
    while have_all_frames do
        -- NOTE:  We rely on all input pipes using the same sample rate!
        --        It doesn't matter which input frame we use for
        --        obtaining frame count.
        local count = self.chan_info[1].in_frame.count

        -- Get input frame pointers.  Note that master channels are
        -- first in the chan_info list.
        for ch_index = 1, #self.chan_info do
            local info = self.chan_info[ch_index]
            if not info.sample_type then
                info.element_size = info.in_frame.element_size
                info.sample_type =
                    utils.frame_data_type(info.element_size)
            end
            info.invals = ffi_cast(info.sample_type, info.in_frame.buf)
        end

        for in_index = 0, count - 1 do
            -- Check each master input sample for an event, storing
            -- samples from both master and slave inputs when an
            -- event is found.
            local have_event = self:check_for_event(in_index)
            if self.skip_first then
                -- Initial sample is considered an event.
                have_event = false
                self.skip_first = false
            end
            if have_event then
                -- Put each master and slaves channel's current input
                -- sample into an output frame.
                for ch_index = 1, self.nbr_output_chans do
                    local info = self.chan_info[ch_index]
                    local outfr = info.out_frame
                    if outfr == nil then
                        --  need new output frame
                        outfr = Frame_create(info.in_frame.maxsize,
                            info.element_size)
                        info.out_frame = outfr
                        outfr.hdr.ts = self.tot_out_sample_count
                        info.outvals =
                            ffi_cast(info.sample_type, outfr.buf)
                        self.out_index = 0
                    end
                    info.outvals[self.out_index] = info.invals[in_index]
                end
                self.out_index = self.out_index + 1
                self.tot_out_sample_count = self.tot_out_sample_count + 1
            end
        end

        for i = 1, self.nbr_output_chans do
            local outfr = self.chan_info[i].out_frame
            -- if we have an output frame put it in output pipe
            if outfr and outfr ~= nil then
                outfr.count = self.out_index
                if not utils.put_frame(outfr,
                        self.chan_info[i].pipe_list) then
                    Frame_delete(outfr)
                end
                self.chan_info[i].out_frame = nil
            end
            -- don't need the input frame any longer
            Frame_delete(self.chan_info[i].in_frame)
        end

        self.tot_count = self.tot_count + count
        -- get a new set of input frames to process
        have_all_frames = self:get_frames()
    end
end

local m = {}

-- Creates a new instantiation of this datamode type.
-- Standard parameters are:
--   stanza --> decoded JSON structure containing the metadata
--       for an instance of this module
--   node_list --> list of all nodes (i.e., module instances, proxy
--       nodes) created to this point, indexed by channel name
--   first_sie_index --> NOT USED
--   meta_list --> list of all created-to-this-point nodes' stanzas,
--       indexed by channel name
-- Return values:
--   the new instance object
function m:new(stanza, node_list, first_sie_index, meta_list)
    assert(stanza, "Missing parameter 'stanza'")
    assert(node_list, "Missing parameter 'node_list'")
    assert(first_sie_index, "Missing parameter 'first_sie_index'")
    assert(meta_list, "Missing parameter 'meta_list'")
    local es_obj = setmetatable({}, {__index = mt})
    es_obj.stanza = stanza
    -- save a needed system preference
    es_obj.prefix_dm_name = stanza.summary.include_dm_name_and_at
    -- don't need to call set_common_meta() since there are
    -- no modules downstream
    es_obj.chan_info = {}
    -- validate and process needed metadata
    local input_channels = stanza.input_channels
    assert(#input_channels >= 1, "Event Slice data mode " ..
        "must have at least 1 input channel")
    local master_channels = stanza.master_input_channels
    assert(#master_channels >= 1, "Event Slice data mode " ..
        "must have at least 1 master channel")
    es_obj.nbr_master_chans = #master_channels
    local slave_channels = stanza.slave_input_channels or {}
    -- Verify that channels are not duplicated in within the
    -- the tables input_channels, master_channels, and
    -- slave_channels.
    local sorted_input_channels = {}
    for i = 1, #input_channels do
        sorted_input_channels[i] = input_channels[i]
    end
    table.sort(sorted_input_channels)
    for i = 1, #input_channels - 1 do
        assert(sorted_input_channels[i] ~= sorted_input_channels[i + 1],
            "Duplicate channels not allowed in 'input_channels'")
    end
    local sorted_master_channels = {}
    for i = 1, #master_channels do
        sorted_master_channels[i] = master_channels[i]
    end
    table.sort(sorted_master_channels)
    for i = 1, #master_channels - 1 do
        assert(sorted_master_channels[i] ~= sorted_master_channels[i + 1],
            "Duplicate channels not allowed in 'master_channels'")
    end
    local sorted_slave_channels = {}
    for i = 1, #slave_channels do
        sorted_slave_channels[i] = slave_channels[i]
    end
    table.sort(sorted_slave_channels)
    for i = 1, #slave_channels - 1 do
        assert(sorted_slave_channels[i] ~= sorted_slave_channels[i + 1],
            "Duplicate channels not allowed in 'slave_channels'")
    end
    es_obj.nbr_output_chans = #master_channels + #slave_channels
    assert(es_obj.nbr_output_chans == #input_channels,
        "Event Slice data mode input channels list must be " ..
        "same size as master channel list plus slave channel list")
    --     verify that a channel appears exactly once in
    --     input_channels table and exactly once in either
    --     the master_channels or slave_channels table.
    local sm_index, ss_index = 1, 1
    for i = 1, #input_channels do
        if sorted_input_channels[i] == sorted_master_channels[sm_index] then
            sm_index = sm_index + 1
        elseif sorted_input_channels[i] == sorted_slave_channels[ss_index] then
            ss_index = ss_index + 1
        else
            assert(false, string.format("input channel '%s' " ..
                "not listed in master or slave channel lists",
                tostring(sorted_input_channels[i])))
        end
    end
    --     node name
    es_obj.name = stanza.name
    assert(es_obj.name, "Missing node name")
    --     combine master channel(s) and slave channel(s) names into
    --     one list with the master channel(s) first
    es_obj.all_input_chans = {}
    for i = 1, #master_channels do
        es_obj.all_input_chans[i] = master_channels[i]
    end
    for i = 1, #slave_channels do
        es_obj.all_input_chans[#es_obj.all_input_chans + 1] =
            slave_channels[i]
    end
    --     sample rate (get and check for consistency)
    for i = 1, #es_obj.all_input_chans do
        local node_meta = meta_list[es_obj.all_input_chans[i]]
        assert(node_meta, "metadata stanza for Input Channel \"" ..
            tostring(es_obj.all_input_chans[i]) .. "\" not found")
        if not es_obj.sample_rate then
            es_obj.sample_rate = node_meta.sample_rate
        else
            assert(es_obj.sample_rate == node_meta.sample_rate,
                tostring(es_obj.all_input_chans[i]) .. " sample rate " ..
                "does not match")
        end
    end
    es_obj.sample_rate = rational:new(es_obj.sample_rate):tonumber()
    assert(es_obj.sample_rate and es_obj.sample_rate > 0,
        string.format("Invalid input sample rate: %s",
        tostring(es_obj.sample_rate)))
    --     "store initial state" flag
    local store_initial = stanza.store_initial_values
    if store_initial and type(store_initial) ~= "boolean" then
        store_initial = stanza.store_initial_values == "true" or false
    end
    es_obj.skip_first = not store_initial
    --     trigger
    local trigger_mode = stanza.trigger_mode
    assert(trigger_mode, "Missing trigger mode")
    --     data_type
    --     (create master input channel info objects to hold
    --      data_type and other channel-specific values)
    for i = 1, #master_channels do
        local node_meta = meta_list[master_channels[i]]
        assert(node_meta, "metadata stanza for Input Channel \"" ..
            tostring(master_channels[i]) .. "\" not found")
        assert(node_meta.data_type, "Missing data_type for \"" ..
            tostring(master_channels[i]) .. "\"")
        es_obj.chan_info[i] =
            {name = master_channels[i],
             stanza = node_meta,
             data_type = node_meta.data_type,
             sie_buf_info = {
                 idx = stanza.sie_buf_info.idx + i - 1,
                 id = stanza.sie_buf_info.id + i - 1,
                 group = stanza.sie_buf_info.group + i - 1,
             },
             element_size = nil,
             sample_type = nil,
             prev_state = 0}
    end
    --     (create slave input channel info objects to hold
    --      data_type and other channel-specific values)
    for i = 1, #slave_channels do
        local node_meta = meta_list[slave_channels[i]]
        assert(node_meta, "metadata stanza for Input Channel \"" ..
            tostring(slave_channels[i]) .. "\" not found")
        assert(node_meta.data_type, "Missing data_type for \"" ..
            tostring(slave_channels[i]) .. "\"")
        es_obj.chan_info[#master_channels + i] =
            {name = slave_channels[i],
             stanza = node_meta,
             data_type = node_meta.data_type,
             sie_buf_info = {
                 idx = stanza.sie_buf_info.idx + i + #master_channels - 1,
                 id = stanza.sie_buf_info.id + i + #master_channels - 1,
                 group = stanza.sie_buf_info.group + i + #master_channels - 1,
             },
             element_size = nil,
             sample_type = nil}
    end
    -- create input pipe structures
    if trigger_mode == "always_on" then
        --     create non-trigger input pipe structures
        local input_node
        local node_meta
        for i = 1, #master_channels do
            input_node = node_list[master_channels[i]]
            assert(input_node, "Node for Input Channel \"" ..
                tostring(master_channels[i]) .. "\" not found")
            es_obj.chan_info[i].in_pipe =
                 Pipe_attach_pipe(input_node.pipe_list)
            assert(es_obj.chan_info[i].in_pipe and
                es_obj.chan_info[i].in_pipe ~= nil,
                "could not create input pipe")
            es_obj.chan_info[i].pipe_list = Pipe_new_pipe_list()
            assert(es_obj.chan_info[i].pipe_list and
                es_obj.chan_info[i].pipe_list ~= nil,
                "could not create pipe list")
        end
        for i = 1, #slave_channels do
            input_node = node_list[slave_channels[i]]
            assert(input_node, "Node for Input Channel \"" ..
                tostring(slave_channels[i]) .. "\" not found")
            es_obj.chan_info[#master_channels + i].in_pipe =
                 Pipe_attach_pipe(input_node.pipe_list)
            assert(es_obj.chan_info[#master_channels + i].in_pipe and
                es_obj.chan_info[#master_channels + i].in_pipe ~= nil,
                "could not create input pipe")
            es_obj.chan_info[#master_channels + i].pipe_list =
                Pipe_new_pipe_list()
            assert(es_obj.chan_info[#master_channels + i].pipe_list and
                es_obj.chan_info[#master_channels + i].pipe_list ~= nil,
                "could not create pipe list")
        end
    else
        --     create trigger module instance
        local trig_chan = stanza.trigger_channel
        assert(trig_chan, "Trigger mode requires a trigger channel")
        local ok, mod =
            pcall(require, 'test_engine/modules/trigger/' .. trigger_mode)
        if not ok then
            error(string.format(__'Could not load trigger ' ..
                 'module %s (%s)', stanza.trigger_mode, stanza.name))
        end
        local trigger_module =
            mod:new(es_obj.all_input_chans, trig_chan, node_list, meta_list)
        if not trigger_module then
            error(string.format(__'Unable to create type %s (%s)',
                'test_engine/modules/trigger/' .. trigger_mode,
                stanza.name))
        end
        assert(#trigger_module.chan_info ==
               #es_obj.all_input_chans, "Number of trigger output pipes " ..
               "does not equal number of input channels")
        --     create input pipe structures using
        --     trigger module output pipes as this instance's
        --     input pipes
        for i = 1, #master_channels do
            es_obj.chan_info[i].in_pipe =
                Pipe_attach_pipe(trigger_module.chan_info[i].pipe_list)
            assert(es_obj.chan_info[i].in_pipe and
                es_obj.chan_info[i].in_pipe ~= nil,
                "could not create input pipe")
            es_obj.chan_info[i].pipe_list = Pipe_new_pipe_list()
            assert(es_obj.chan_info[i].pipe_list and
                es_obj.chan_info[i].pipe_list ~= nil,
                "could not create pipe list")
        end
        for i = 1, #slave_channels do
            es_obj.chan_info[#master_channels + i].in_pipe =
                Pipe_attach_pipe(
                    trigger_module.chan_info[#master_channels + i].pipe_list)
            assert(es_obj.chan_info[#master_channels + i].in_pipe and
                es_obj.chan_info[#master_channels + i].in_pipe ~= nil,
                "could not create input pipe")
            es_obj.chan_info[#master_channels + i].pipe_list =
                Pipe_new_pipe_list()
            assert(es_obj.chan_info[#master_channels + i].pipe_list and
                es_obj.chan_info[#master_channels + i].pipe_list ~= nil,
                "could not create pipe list")
        end
        es_obj.trigger_module = trigger_module
    end

    es_obj.out_index = 0
    -- sample counts for this module
    es_obj.tot_out_sample_count = 0
    es_obj.tot_count = 0

    -- second returned value will indicate one sie buf per input
    -- channel
    return es_obj, es_obj.nbr_output_chans
end

return m

