require 'strict'

-- peak_valley_matrix.lua - Implements a one or two dimensional histogram
-- data mode which stores peak-valley data.
--
-- This module comprises two main submodules.  The first processes a
-- channel's input frames to produce Peak Valley data, which is given
-- to the second to produce the output for a histogram.

local cjson = require 'cjson'
local ffi = require 'ffi'
local C = ffi.C 
local Pipe = require 'test_engine/pipe'
local Frame = require 'test_engine/frame'
local hg = require 'test_engine/modules/common/histogram'
local pv = require 'test_engine/modules/common/pv_algo'
local rational = require 'rational'
local sie_schema = require 'test_engine/sie_schema'
local sie_utils = require 'test_engine/sie_utils'
local utils = require 'test_engine/test_eng_utils'

local ffi_cast = ffi.cast
local ffi_copy = ffi.copy
local ffi_sizeof = ffi.sizeof
local ffi_string = ffi.string

local Pipe_attach_pipe = Pipe.attach_pipe
local Pipe_get_frame = Pipe.get_frame
local Pipe_has_frames = Pipe.has_frames
local Frame_delete = Frame.delete

local math_floor = math.floor
local math_min = math.min
local math_max = math.max

local ctor_cache = {}

-- mh stores histogram objects information
local mh = {}
-- mpv stores peak valley objects information
local mpv = {}

local config = require 'system_config'
local libte = ffi.load(config.test_engine_library)
local libteFlushFrames = libte.SIEBuffer_flush_frames
local libteFlushThese = libte.SIEWriter_flush_these
local libteFlush = libte.SIEBuffer_flush
local libteAmountWritten = libte.SIEWriter_amount_written
local libteAppend = libte.SIEBuffer_append
local libteEnq = libte.SIEBuffer_enq


--=== Histograms Object's definition

-- local "constants"
local HIST_TYPE_RANGE_ONLY = 1
local HIST_TYPE_RANGE_MEAN = 2
local HIST_TYPE_TO_FROM    = 3

-- function mh:run_evenly_divided_sequential(in_buf, count)
-- function mh:run_evenly_divided_message(in_buf, count)

-- Computes peak-to-valley and valley-to-peak ranges and
-- histograms the results.
-- Parameters:
--   hist    -  individual histogram object
--   pvdata  -  array containing peak-valley data
--   hdata   -  array to use in calculating values to histogram
function mh:run_range_only(hist, pvdata, hdata)
    local bin
    local dlen = 0

    for i = 1, pvdata.count do
        if hist.first_time == 1 then
            hist.first_time = 0
        else
            hdata[dlen] = math.abs(pvdata[i] - hist.prev_val)
            dlen = dlen + 1
        end
        hist.prev_val = pvdata[i]
    end

    if dlen > 0 then
        hg.histogram(hist, hdata, dlen)
    end
end

-- Computes peak-to-valley and valley-to-peak ranges and
-- histograms the results along with the average value of
-- the peak-valley/valley-peak pairs.
-- Parameters:
--   hist    -  individual histogram object
--   pvdata  -  array containing peak-valley data
--   hdata   -  array to use in calculating values to histogram
function mh:run_range_mean(hist, pvdata, hdata)
    local bin
    local dlen = 0

    for i = 1, pvdata.count do
        if hist.first_time == 1 then
            hist.first_time = 0
        else
            hdata[dlen] = math.abs(pvdata[i] - hist.prev_val)
            dlen = dlen + 1
            hdata[dlen] = (pvdata[i] + hist.prev_val) / 2
            dlen = dlen + 1
        end
        hist.prev_val = pvdata[i]
    end

    if dlen > 0 then
        hg.histogram(hist, hdata, dlen)
    end
end

-- Histograms peak-valley/valley-peak pairs by storing the first
-- (with respect to time) value of the pair followed by the second
-- value.
-- Parameters:
--   hist    -  individual histogram object
--   pvdata  -  array containing peak-valley data
--   hdata   -  array to use in calculating values to histogram
function mh:run_to_from(hist, pvdata, hdata)
    local bin
    local dlen = 0

    for i = 1, pvdata.count do
        if hist.first_time == 1 then
            hist.first_time = 0
        else
            hdata[dlen] = hist.prev_val
            dlen = dlen + 1
            hdata[dlen] = pvdata[i]
            dlen = dlen + 1
        end
        hist.prev_val = pvdata[i]
    end

    if dlen > 0 then
        hg.histogram(hist, hdata, dlen)
    end
end

-- Places the given arrays of peak-valley data in the appropriate
-- bins of their associated histograms.
-- Parameters:
--   pv_arrays  - arrays containing peak-valley data
--   hdata      -  array to use in calculating values to histogram
function mh:run(pv_arrays, hdata)
    for i = 0, self.num_hist - 1 do
        local hist = self.hist[i]
        if hist.histogram_code == HIST_TYPE_RANGE_ONLY then
            self:run_range_only(hist, pv_arrays[i + 1], hdata)
        elseif hist.histogram_code == HIST_TYPE_RANGE_MEAN  then
            self:run_range_mean(hist, pv_arrays[i + 1], hdata)
        elseif hist.histogram_code == HIST_TYPE_TO_FROM then
            self:run_to_from(hist, pv_arrays[i + 1], hdata)
        end
    end
end

-- At the end of the test, writes to the SIE file the bins data
-- for the histograms.
-- Parameters:
--   sie_writer  -  SIE file writer object
--   pv_obj      -  the Peak Valley algorithm object
--   hdata       -  array to use in calculating values to histogram
function mh:finalize(sie_writer, pv_obj, hdata)
    if self.final_called ~= 0 then return end
    self.final_called = 1
    local pv_arrays = pv_obj:finalize()
    if pv_arrays then
        -- process final peak-valley values
        self:run(pv_arrays, hdata)
    end
    local num = self.num_hist - 1
    local bin_bits = 64
    local hdr = ffi.new('uint32_t[3]')
    for i = 0, num do
        local hist = self.hist[i]
        local buf = sie_writer.bufs[hist.sie_idx]
        hdr[0] = hist.nbr_dims 
        hdr[1] = hist.total_dimension_size
        if hist.nbr_dims > 1 then
            hdr[2] = hist.total_dimension_size
            libteEnq(buf, hdr, ffi_sizeof('uint32_t') * 3)
        else
            libteEnq(buf, hdr, ffi_sizeof('uint32_t') * 2)
        end
        libteEnq(buf, hist.binmatrix, hist.matrix_size * (bin_bits / 8))
        libteFlush(buf, 1)
    end
end

-- Writes a histogram's bounds information to the SIE file using
-- sie_writer's "scratch" sie buf.
-- Parameter:
--   sw    -  SIE file writer object
--   hist  -  array of histogram tables
--   bounds is -  bounds table
function mh:write_bounds(sw, hist, bounds_id)
    local dim_size = hist.total_dimension_size
    local bounds = hist.dims[bounds_id].bounds

    local bounds_bits = 64  -- all bounds use 64 bits (for now!)
    sw:new_buffer(sw.scratch_buf_idx, hist.group_id + bounds_id,
        (dim_size + 1) * bounds_bits / 8, bounds_bits / 8, 0)
    local bbuf = sw.bufs[sw.scratch_buf_idx]
    return bbuf.group
end

function mh:flush_bounds(sw, dim_size, bounds)
    local bounds_bits = 64  -- all bounds use 64 bits (for now!)
    local ptr = ffi.cast('const char *', bounds)
    local bbuf = sw.bufs[sw.scratch_buf_idx]
    libteAppend(bbuf, ptr, (dim_size + 1) * bounds_bits / 8)
    libteFlush(bbuf, 0)
end

-- Create all of the SIE buffers needed by this sink
function mh:setup()
    -- nothing yet
end

-- provides definition of a histogram table
local function meta_ctor(t)
    if not ctor_cache[t] then
        local hdm_type = string.format('struct %s_pvm_hdm', t)
        local spec = string.gsub([[
            struct pvm_dimension {
                double          lo_bound;
                double          hi_bound;
                double          base;
                double          bin_size;
                int             dimension_size;
                int             total_dimension_size;
                double          *bounds;
            };
            struct $t_pvm_hist {
                int             nbr_dims;
                int             dimension_size;
                int             total_dimension_size;
                int             histogram_code;
                int             matrix_size;
                struct pvm_dimension *dims;
                $t              *binmatrix;
                /* For evenly_divided special case */
                double          prev_val;
                int             sie_idx;
                char            first_time;
                char            *name;
                char            *description;
                char            *hist_name;
                char            *category;
                char            *units;
                double          sample_rate;
                int             ch_id;
                int             group_id;
            };
            struct $t_pvm_hdm {
                int             num_hist;
                int             final_called;
                struct $t_pvm_hist  *hist;
            };
        ]], '%$(%w+)', {t = t})
        ffi.cdef(spec)
        ctor_cache[t] = ffi.metatype(hdm_type, {__index = mh})
    end
    return ctor_cache[t]
end

--=== Peak Valley object's definition

-- Finishes initialization of this object.
function mpv:setup()
    self.pv_arrays = {}
    for i = 1, #self.chan_info do
        self.pv_arrays[i] = {}
        self.pv_arrays[i].count = 0
    end
end

-- Gets a frame from each input pipe, storing the frames in
-- the input pipe's 'self.chan_info[<pipe index>].in_frame' entry.
-- Returns:  true if a full set of frames is acquired,
--           false otherwise
function mpv:get_frames()
    -- check that each input pipe has a frame, so we don't
    -- "get" a frame from any pipe which we can't use
    local got_frames = #self.chan_info > 0
    if not got_frames then
        return false  -- short circuit return!
    end

    for i = 1, #self.chan_info do
        got_frames = Pipe_has_frames(self.chan_info[i].in_pipe)
        if not got_frames then
            break  -- don't have a frame from this input pipe
        end
    end
    if got_frames then
        for i = 1, #self.chan_info do
            self.chan_info[i].in_frame = Pipe_get_frame(self.chan_info[i].in_pipe)
        end
    else
        -- ensure we don't accidentally use old data
        for i = 1, #self.chan_info do
            self.chan_info[i].in_frame = nil
        end
    end
    return got_frames
end

-- Computes the peaks and valleys for the input channels, one
-- full set of incoming frames at a time.
function mpv:run()
    -- attempt to get a full set of input frames (one from each
    -- input pipe)
    local have_all_frames = self:get_frames()
    if not have_all_frames then
        -- can't process any input frames yet
        return nil
    end

    -- Process each frame in input pipe(s).
    while have_all_frames do
        -- NOTE:  We rely on all input pipes using the same sample rate!
        --        It doesn't matter which input frame we use for
        --        obtaining frame count.
        local count = self.chan_info[1].in_frame.count

        -- Process frame for each input channel, setting input frame
        -- pointers, creating output frames as needed (with output
        -- frame pointers).
        for i = 1, #self.chan_info do
            local infr = self.chan_info[i].in_frame
            if not self.chan_info[i].sample_type then
                self.chan_info[i].sample_type =
                    utils.frame_data_type(infr.element_size)
            end
            self.chan_info[i].invals = ffi_cast(self.chan_info[i].sample_type, infr.buf)

            -- find pv's (peaks and valleys)
            pv.run_pv_algo(count, self.chan_info[i],
                           self.chan_info[i].hysteresis, self.pv_arrays[i])
            self.chan_info[i].last_input_value =
                self.chan_info[i].invals[count - 1]
        end

        -- don't need the processed input frames any longer
        for i = 1, #self.chan_info do
            Frame_delete(self.chan_info[i].in_frame)
        end

        -- get a new set of input frames to process
        have_all_frames = self:get_frames(self.chan_info)
    end

    return self.pv_arrays
end

-- Finishes processing of outgoing frames at end of test.
-- NOTE:  Any last data to be written to the SIE file must
--        be written here, since no module's run() function
--        will be called again!
-- Parameter:
--     'sie_writer'  SIE writer
function mpv:finalize()
    -- return immediately if there isn't enough information
    -- to process a "last value"
    for i = 1, #self.chan_info do
        local info = self.chan_info[i]
        if not info.last_input_value then
            return
        end
    end

    for i = 1, #self.chan_info do
        local info = self.chan_info[i]
        if info.state == "initial" then
            self.pv_arrays[i][1] = info.last_input_value
            self.pv_arrays[i].count = 1
        elseif info.state == "valley_search" then
            if info.valley == info.last_input_value or
               info.last_input_value - info.valley < info.hysteresis then
                self.pv_arrays[i][1] = info.valley
                self.pv_arrays[i].count = 1
            end
        else  -- info.state == "peak_search"
            if info.peak == info.last_input_value or
               info.peak - info.last_input_value < info.hysteresis then
                self.pv_arrays[i][1] = info.peak
                self.pv_arrays[i].count = 1
            end
        end
    end

    return self.pv_arrays
end

--=== Peak Valley Matrix object's definition

require 'middleclass'

local m = class('peak_valley_matrix')

-- Creates the SIE buffers (pipes and XML) required.
-- There will be multiple channel descriptions per input channel,
-- and a generic histogram lookup table description if it doesn't
-- already exist.
function m:sie_xml(sie_writer, dm_tab, stanza_list)
    for i = 0, self.num_hist - 1 do
        local hist = self.hdm.hist[i]

        -- get channel id for the appropriate generic histogram
        -- lookup table channel, writing it if necessary
        local bin_tab = {}
        bin_tab._base_id =
            sie_utils.ensure_sie_bounds_channel(sie_writer, 64, hist.ch_id)
        bin_tab._dm = ffi.string(hist.hist_name)
        bin_tab._name = ffi.string(hist.name)
        -- write "bin" min/max values and the histogram "bin"
        -- channels which will use them
        local index_chan_id = {}
        -- for dimension 0
        --     write bounds information to the SIE file
        -- [the following call prevents the bounds from being written
        --  twice - both before and after the dim bounds channel
        --  DON'T KNOW WHY, so FIXME]
        sie_writer:xml()

        -- Generate the bounds 0 information data block
        bin_tab._sie_group = self.hdm:write_bounds(sie_writer, hist, 0)
        bin_tab._histdim = 0
        bin_tab._sie_ch = hist.ch_id + 1

        -- Write the bounds 0 information XML
        index_chan_id[0] = sie_utils.write_bin_channel(sie_writer, bin_tab)

        -- Write the bounds 0 information data block
        self.hdm:flush_bounds(sie_writer, hist.total_dimension_size,
            hist.dims[0].bounds)

        if hist.nbr_dims > 1 then
            -- for dimension 1
            --     write bounds 1 information to the SIE file
            -- Generate the bounds 1 information data block
            bin_tab._sie_group = self.hdm:write_bounds(sie_writer, hist, 1)

            -- Write the dim 1 channel which will use the bounds info
            bin_tab._histdim = 1
            bin_tab._sie_ch = hist.ch_id + 2
            -- Write the bounds 1 information XML
            index_chan_id[1] = sie_utils.write_bin_channel(sie_writer, bin_tab)

            -- Write the bounds 2 information data block
            self.hdm:flush_bounds(sie_writer, hist.total_dimension_size,
                hist.dims[1].bounds)
        end
        -- write XML for the histogram's SIE channel
        local sch_tab = {}
        sch_tab.description = ffi.string(hist.description) or ""
        sch_tab.datamode_name = ffi.string(hist.hist_name)
        if self.prefix_dm_name then
            sch_tab.sie_name = sch_tab.datamode_name .. '@' ..
                ffi.string(hist.name)
        else
            sch_tab.sie_name = ffi.string(hist.name)
        end
        sch_tab.somat_input_channel = ffi.string(hist.name)
        sch_tab.category = ffi.string(hist.category)
        sch_tab.data_type = dm_tab.data_type
        sch_tab.datamode_type = dm_tab.datamode_type
        sch_tab.sample_rate = hist.sample_rate
        sch_tab.nbr_data_bits = 64
        sch_tab.name = ffi_string(hist.name)
        sch_tab.units = ffi_string(hist.units)
        sch_tab.dim1_name =
            (self.histogram_code == HIST_TYPE_RANGE_ONLY or
             self.histogram_code == HIST_TYPE_RANGE_MEAN) and
            ("Range " .. sch_tab.name) or ("From " .. sch_tab.name)
        sch_tab.dim2_name =
            (self.histogram_code == HIST_TYPE_RANGE_MEAN and
             ("Mean " .. sch_tab.name)) or
            (self.histogram_code == HIST_TYPE_TO_FROM and
             ("To " .. sch_tab.name)) or "Cycles"
        sch_tab.nbr_dims = hist.nbr_dims
        sch_tab.low_bound = {}
        sch_tab.high_bound = {}
        sch_tab.low_bound[0] = hist.dims[0].lo_bound
        sch_tab.high_bound[0] = hist.dims[0].hi_bound
        if hist.nbr_dims > 1 then
            sch_tab.low_bound[1] = hist.dims[1].lo_bound
            sch_tab.high_bound[1] = hist.dims[1].hi_bound
        end
        local schema_name = sie_schema.get_dm_schema_name(sch_tab.category)
        local schema_info = sie_schema.get_schema_info(schema_name, sch_tab)
        -- assign index channel ids and dim labels
        schema_info.dims[2]._index_ch = index_chan_id[0]
        schema_info.dims[3]._index_ch = index_chan_id[0]
        if hist.nbr_dims > 1 then
            schema_info.dims[4]._index_ch = index_chan_id[1]
            schema_info.dims[5]._index_ch = index_chan_id[1]
        end
        local input_stanza = stanza_list[ffi_string(hist.name)]
        sie_utils.supplement_stanza_info(
            input_stanza, ffi.string(hist.hist_name))

        input_stanza.sie_buf_info = {
            id = hist.ch_id + hist.nbr_dims + 1,
            group = hist.group_id + hist.nbr_dims,
        }

        local md = sie_utils.write_sie_channel(sie_writer, schema_info,
            input_stanza)
        -- Create the SIE buffer and attach the pipe to it
        local bin_bits = 64
        sie_writer:new_buffer(hist.sie_idx, hist.group_id + hist.nbr_dims,
            hist.matrix_size * bin_bits / 8, bin_bits / 8,
            ffi_sizeof('uint64_t') * (hist.nbr_dims + 1))

        -- Ensure there is file space to write the channel data
        -- at end of test.
        sie_writer:reserve_space(
            ((hist.matrix_size + 2) * (bin_bits / 8)) +   -- bin space
            (ffi_sizeof('uint32_t') * 2) + -- header space
            (ffi_sizeof('uint32_t') * 5), -- SIE overheader
            function()
                self.hdm:finalize(sie_writer, self.pv_obj, self.hdata)
            end)
    end
end

function m:setup(sie_writer, is_restart, stanza_list)
    self:sie_xml(sie_writer, self.dm_tab, stanza_list)
    self.hdata = {}
    self.hdm:setup()
    self.pv_obj:setup()
end

function m:run()
    if self.trigger_module then self.trigger_module:run() end
    -- pv_obj processes input frames
    local pv_arrays = self.pv_obj:run()
    if pv_arrays then
        self.hdm:run(pv_arrays, self.hdata)
    end
end

--[[
function m:finalize(sie_writer)
    self.hdm:finalize(sie_writer)
end
]]

function m:initialize(stanza, node_list, sie_idx, stanza_list)
    local bin_type = stanza.bin_type or 'uint64_t'
    local ctor = meta_ctor(bin_type)

    --== initialize module-level items
    self.final_called = 0
    local input_channels = stanza.input_channels
    self.sie_buffers_needed = #input_channels
    self.nbr_input_channels = #input_channels
    -- If triggered then instantiate a new trigger module
    self.trigger_module = stanza.trigger_mode ~= 'always_on' and
        require('test_engine/modules/trigger/' ..
                stanza.trigger_mode):
                    new(stanza.input_channels, stanza.trigger_channel,
                        node_list, stanza_list)
    -- save a needed system preference
    self.prefix_dm_name = stanza.summary.include_dm_name_and_at

    --== initialize peak-valley support

    self.pv_obj = setmetatable({}, {__index = mpv})
    assert(stanza.pv_hysteresis, "Missing hysteresis")
    local hyster_val, hyster_percent =
        tostring(stanza.pv_hysteresis):match("([%d%.]+)(%%?)")
    hyster_val = rational:new(hyster_val):tonumber()
    if hyster_percent == "%" then
        hyster_percent = hyster_val
        assert(hyster_percent and (hyster_percent >= 0.0) and
               (hyster_percent <= 100.0) and (type(hyster_percent) == 'number'),
               "Invalid hysteresis percent")
        hyster_percent = hyster_percent / 100.0
    else
        hyster_percent = nil
        assert(hyster_val and (hyster_val >= 0.0) and
               (type(hyster_val) == 'number'), "Invalid hysteresis value")
    end
    self.pv_obj.chan_info = {}
    for i = 1, #input_channels do
        local ic = input_channels[i]
        local ic_meta = stanza_list[ic]
        local ic_node = node_list[ic]
        assert(ic_meta, "metadata stanza for Input Channel \"" ..
            tostring(input_channels[i]) .. "\" not found")
        assert(ic_meta.data_type, "Missing data_type for \"" ..
            tostring(input_channels[i]) .. "\"")
        self.pv_obj.chan_info[i] =
            {name = ic,
             stanza = ic_meta,
             data_type = ic_meta.data_type,
             sample_rate = rational(ic_meta.sample_rate):tonumber(),
             sample_type = nil,
             hysteresis = nil,
             state = "initial",
             nbr_input_channels = self.nbr_input_channels,
             valley = math.huge,
             peak = -(math.huge - 1),
             last_input_value = nil}
        if hyster_percent then
            local range_min = tonumber(ic_meta.range_min)
            assert(type(range_min) == "number", "Missing or invalid range_min")
            local range_max = tonumber(ic_meta.range_max)
            assert(type(range_max) == "number", "Missing or invalid range_max")
            assert(range_max > range_min)
            self.pv_obj.chan_info[i].hysteresis =
               (range_max - range_min) * hyster_percent
        else
            self.pv_obj.chan_info[i].hysteresis = hyster_val
        end

        if not self.trigger_module then
            -- Not triggered...attach directly to input channel
            self.pv_obj.chan_info[i].in_pipe = Pipe_attach_pipe(ic_node.pipe_list)
        else
            -- Triggered...attach to gated channel
            self.pv_obj.chan_info[i].in_pipe =
                Pipe_attach_pipe(self.trigger_module.chan_info[i].pipe_list)
        end
    end

    --== initialize histogram support

    self.num_hist = #input_channels
    self.dm_tab = {}
    self.dm_tab.data_type = ""  -- data_type irrelevant for histogram
    self.histogram_type = stanza.histogram_type
    assert(self.histogram_type == "histogram_range_only" or
           self.histogram_type == "histogram_range_mean" or
           self.histogram_type == "histogram_to_from",
        "Invalid histogram type")
    local nbr_dims
    if self.histogram_type == "histogram_range_only" then
        self.dm_tab.datamode_type = "peak_valley_matrix/range_only"
        nbr_dims = 1
        self.histogram_code = HIST_TYPE_RANGE_ONLY
    elseif self.histogram_type == "histogram_range_mean" then
        self.dm_tab.datamode_type = "peak_valley_matrix/range_mean"
        nbr_dims = 2
        self.histogram_code = HIST_TYPE_RANGE_MEAN
    else
        self.dm_tab.datamode_type = "peak_valley_matrix/from_to"
        nbr_dims = 2
        self.histogram_code = HIST_TYPE_TO_FROM
    end

    self.hdm = ctor()  -- named "hdm" for "histogram data mode"
    -- all dimensions have the same number of bins (add 2 extra bins
    -- for underflow and overflow)
    self.hdm.num_hist = self.num_hist
    self.hdm.hist = ffi_cast(string.format('struct %s_pvm_hist *', bin_type),
        C.calloc(self.hdm.num_hist,
        ffi_sizeof(string.format('struct %s_pvm_hist', bin_type))))
    assert(self.hdm.hist ~= 0, "Allocation failed: self.hdm.hist")
    local objs = self.hdm.hist

    for i = 0, self.num_hist - 1 do
        local hb = stanza.histogram_bins[i + 1]
        local obj = objs[i]
        local ic = stanza.input_channels[i + 1]
        local chan_meta = stanza_list[ic]

        assert(hb.type == "evenly_divided",
                "Only 'evenly_divided' is supported")
        obj.first_time = 1
        obj.prev_val = 0
        obj.sie_idx = sie_idx + i
        obj.ch_id = stanza.sie_buf_info.id +
            i * (stanza.histogram_type == "histogram_range_only" and 2 or 3)
        obj.group_id = stanza.sie_buf_info.group +
            i * (stanza.histogram_type == "histogram_range_only" and 2 or 3)

        obj.sample_rate = rational(chan_meta.sample_rate):tonumber()
        obj.nbr_dims = nbr_dims
        local units = chan_meta.output_units or ''
        obj.units = ffi_cast('char *', C.calloc(units:len() + 1,
                ffi_sizeof('char')))
        assert(obj.units ~= nil, "Allocation failed: obj.units")
        ffi_copy(obj.units, units)
        -- allocate space for histogram's bin matrix 
        obj.dimension_size = stanza.number_of_bins
        obj.total_dimension_size = obj.dimension_size + 2
        obj.matrix_size = obj.total_dimension_size ^ obj.nbr_dims
        obj.binmatrix = ffi_cast(string.format('%s *', bin_type),
                C.calloc(obj.matrix_size, ffi_sizeof(bin_type)))
        assert(obj.binmatrix ~= nil, "Allocation failed: obj.binmatrix")
        -- initialize dimensions information
        obj.dims = ffi_cast('struct pvm_dimension *',
                C.calloc(obj.nbr_dims, ffi_sizeof('struct pvm_dimension')))
        assert(obj.dims ~= nil, "Allocation failed: obj.dims")
        for j = 0, obj.nbr_dims - 1 do  -- nbr_dims == 1 or 2 for now
            obj.dims[j].dimension_size = obj.dimension_size
            obj.dims[j].total_dimension_size = obj.total_dimension_size
            obj.dims[j].bin_size =
                    (hb.high_bound - hb.low_bound) / obj.dims[j].dimension_size
            obj.dims[j].bounds = ffi_cast('double *',
                C.calloc(obj.total_dimension_size + 1, ffi_sizeof('double')))
            assert(obj.dims[j].bounds ~= nil,
                "Allocation failed: obj.dims[" .. tostring(j) .. "].bounds")
            obj.histogram_code = self.histogram_code
            -- all bins are to be evenly divided
            if j == 0 and (self.histogram_code == HIST_TYPE_RANGE_ONLY or
                           self.histogram_code == HIST_TYPE_RANGE_MEAN) then
                obj.dims[j].lo_bound = 0
                obj.dims[j].hi_bound = hb.high_bound - hb.low_bound
            else
                obj.dims[j].lo_bound = hb.low_bound
                obj.dims[j].hi_bound = hb.high_bound
            end
            obj.dims[j].base = obj.dims[j].lo_bound - obj.dims[j].bin_size
            obj.dims[j].bounds[0] = -1/0
            obj.dims[j].bounds[obj.total_dimension_size] = 1/0
            for k = 0, obj.total_dimension_size - 2 do
                obj.dims[j].bounds[k + 1] = obj.dims[j].lo_bound
                    + k * obj.dims[j].bin_size
            end
        end
        -- initialize remaining histogram information
        local desc = chan_meta.description or ic or
            'No channel description'
        obj.description = ffi_cast('char *', C.calloc(desc:len() + 1,
            ffi_sizeof('char')))
        assert(obj.description ~= nil, "Allocation failed: obj.description")
        ffi_copy(obj.description, desc)
        obj.name = ffi_cast('char *',
            C.calloc(ic:len() + 1, ffi_sizeof('char')))
        assert(obj.name ~= nil, "Allocation failed: obj.name")
        ffi_copy(obj.name, ic)
        obj.hist_name = ffi_cast('char *',
            C.calloc(stanza.name:len() + 1, ffi_sizeof('char')))
        assert(obj.hist_name ~= nil, "Allocation failed: obj.hist_name")
        ffi_copy(obj.hist_name, stanza.name)
        obj.category = ffi_cast('char *',
            C.calloc(stanza.category:len() + 1, ffi_sizeof('char')))
        assert(obj.category ~= nil, "Allocation failed: obj.category")
        ffi_copy(obj.category, stanza.category)
    end

    return self, #input_channels
end

function m.get_sie_buffer_info(stanza)
    local extra_ch_ids =
        stanza.histogram_type == "histogram_range_only" and 1 or 2

    return #stanza.input_channels,
        #stanza.input_channels * (extra_ch_ids == 1 and 2 or 3) + extra_ch_ids,
        #stanza.input_channels * (extra_ch_ids == 1 and 2 or 3)
end

return m
