local armada = require 'armada'
local ev = require 'ev'
local ffi = require 'ffi'
local C = ffi.C
local ts = require 'timestamps'
local channel_proxy = require 'channel_proxy'
local unix = require 'unix'
local tablex = require 'tablex'
local config = require 'system_config'

local Pipe

local TestProxy = class('TestProxy', channel_proxy)


-- Get the list of channels from the setup that will be provided to the outside
-- world by this proxy.
function TestProxy:_get_proxy_channels(setup)
    local channel_list = {}
    for _, stanza in ipairs(setup) do
        -- Only filter (computed) channels are made available by
        -- test engine proxies
        if stanza.category and stanza.category:find('^filter/') then
            -- Only locally available channels are provided by this proxy
            if stanza.network_node and stanza.network_node == self.host then
                -- Only put things here that should be included in the
                -- Available channels information
                channel_list[#channel_list + 1] = {
                    name = 'remote_' .. stanza.name,
                    sample_rate = stanza.sample_rate,
                    sample_rate_legacy_uncorrected =
                        stanza.sample_rate_legacy_uncorrected,
                    data_type = stanza.data_type,
                    network_node = stanza.network_node,
                    category = 'source/test-engine-export',
                }
            end
        end
    end
    return channel_list
end


-- Function used to read from channel input pipes and write the data to the
-- sink(s)
function TestProxy:_process_pipes()
    for _, chan in ipairs(self.channel_list) do
        -- Is there a pipe list for this channel?
        if chan.pipe_list ~= nil and chan.pipe_list[0] ~= nil then
            -- Yes, are there any frames in it?
            local frame = Pipe.get_frame(chan.pipe_list[0])
            assert(frame.element_size == chan.element_size,
                "Element size does not match data type size")
            while frame ~= nil do
                for _, sink in pairs(chan.channel_sinks) do
                    sink(chan.ts, frame.buf, frame.count * frame.element_size)
                end
                chan.ts = chan.ts + frame.count
                frame = Pipe.get_frame(chan.pipe_list[0])
            end
            Pipe.flush_frames(chan.pipe_list[0])
        end
    end
end

-- Initialize the proxy object
function TestProxy:initialize(hostname, pipe_rate, setup, pipe)
    Pipe = pipe
    self.host = hostname
    self.proxy_type = 'TestProxy'

    -- Get the list of channels that will be made available by this proxy
    self.channel_list = self:_get_proxy_channels(setup)

    -- This list is the list provided to available channels
    self.pristine_channel_list = tablex.deepcopy(self.channel_list)

    -- Now do the base class initialization
    channel_proxy.initialize(self, 'test-engine')

    -- Add additional information needed for the proxies list of channels
    -- that is not provided in available channels information.  E.G. The info
    -- added here is proxy bookeepping information.
    for i, chan in ipairs(self.channel_list) do
        chan.channel_sinks = {}
        chan.pipe_list = Pipe.new_pipe_list()
        chan.connector = self.pristine_channel_list[i].connector
        chan.element_size = tonumber(chan.data_type:match('_float(%d+)')) / 8
    end

    -- Create a timer that is used to service the channel sinks
    local rate = 1.0 / pipe_rate
    ev.Timer.new(function(l, t, e)
            self:_process_pipes()
        end, rate, rate):start(ev.Loop.default)
end


-- Tell the world what channels are available from this proxy
function TestProxy:available_channels(hostname)
    -- Assign the connectors
    for i, ci in ipairs(self.pristine_channel_list) do
        self.pristine_channel_list[i].connector =
            ci.name .. '@' .. self.proxy_hostname
    end

    return self.pristine_channel_list
end


-- No channels are provided on the comet stream
function TestProxy:start_comet_stream(w, chans, sink) end


-- Configure a channel and prepare it to start
function TestProxy:start_channel_stream(w, spec, md_sink, sink)
    for i, chan in ipairs(self.channel_list) do
        if chan.connector == spec.connector then
            chan.ts = 0
            self.start_callback = function(self, start_time)
                    self.start_callback = nil
                end
            chan.channel_sinks[w] = sink
            w.cleanup = function()
                chan.channel_sinks[w] = nil
                if not next(chan.channel_sinks) and chan.pipe_list ~= nil and
                    chan.pipe_list[0] ~= nil then
                    Pipe.flush_frames(chan.pipe_list[0])
                    Pipe.detach_pipe(chan.pipe_list, chan.pipe_list[0])
                end
            end
            md_sink(spec)
            self:announce_ready(spec.connector, true)
        end
    end
end

-- No control functionality
function TestProxy:control(loop, w, http, spec, cb)
    if not spec.request then
        return {
            msg = string.format('Control request not specified on %s',
                                self.proxy_hostname),
            result = 'error',
        }
    end
    if not spec.connector then
        return {
            msg = string.format('Control with no connector specified on %s',
                                self.proxy_hostname),
            result = 'error',
        }
    end
    for _, c in ipairs(self.pristine_channel_list) do
        if c.connector == spec.connector then
            if not self[spec.request] or
                type(self[spec.request]) ~= 'function' then
                return {
                    msg = string.format("Control request %s invalid on %s",
                                        spec.request, c.connector),
                    result = 'error',
                }
            end
            return self[spec.request](self, spec.params)
        end
    end
    return {
        msg = string.format("Control connector %s not found", spec.connector),
        result = 'error',
    }
end

return TestProxy
